require 'omniauth'
require 'jwt'

module OmniAuth
  module Strategies
    class Nightcrawler
      include OmniAuth::Strategy

      class JWTPayloadError < StandardError; end
      class JWTResponseError < StandardError; end

      def request_phase
        redirect request_phase_url
      end

      def callback_phase
        code = request.params["code"]

        res = HTTParty.post(token_url,
          body: token_callback_payload(code),
          headers: {'Content-Type' => 'application/json'}
        )

        if res.response.code == "200"
          issuer_pubkey_str = Base64.decode64 options.issuer_pubkey
          issuer_pubkey = OpenSSL::PKey::RSA.new issuer_pubkey_str

          response = JSON.parse(res.body)
          @raw_token = response["access_token"]

          @decoded = JWT.decode(
            @raw_token,
            issuer_pubkey,
            true, # verify sig
            {algorithm: 'RS256'}
          )[0]["data"]

          (options.required_payload_keys || []).each do |field|
            raise JWTPayloadError.new("Missing required '#{field}' info.") if !@decoded.key?(field.to_s)
          end

          super
        else
          raise JWTResponseError
        end
      rescue JWTPayloadError => e
        fail! :claim_invalid, e
      rescue JWTResponseError => e
        fail! :response_invalid, e
      end

      #####

      option :app_id, nil
      option :app_secret, nil
      option :auth_url_base, nil
      option :issuer_pubkey, nil
      option :required_payload_keys, %w(user_info)

      uid{ @decoded["user_info"]["tusker"] }

      extra do
        { raw_token: @raw_token }
      end

      info { @decoded["user_info"] }

      private

      def issuer_pubkey
        issuer_pubkey_str = Base64.decode64(options.issuer_pubkey)
        OpenSSL::PKey::RSA.new(issuer_pubkey_str)
      end

      def token_callback_payload(code)
        {
          client_id: options.app_id,
          client_secret: options.app_secret,
          redirect_uri: callback_url,
          grant_type: "authorization_code",
          code: code
        }.to_json
      end

      def authorize_url
        "#{options.auth_url_base}/oauth/authorize"
      end

      def token_url
        "#{options.auth_url_base}/oauth/token"
      end

      def refresh_url
        "#{options.auth_url_base}/oauth/refresh"
      end

      def callback_url
        "#{host_string}/auth/nightcrawler/callback"
      end

      def request_phase_url
        "#{authorize_url}?client_id=#{options.app_id}&response_type=code&redirect_uri=#{CGI.escape(callback_url)}"
      end

      def host_string
        "#{request.env['rack.url_scheme']}://#{request.host_with_port}"
      end

    end
  end
end
